(************************************************************************************
Filename    :   Win32_BasicVR.h
Content     :   Core components for achieving basic VR, shared amongst samples
Created     :   October 20th, 2014
Author      :   Tom Heath
Copyright   :   Copyright 2014 Oculus, Inc. All Rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*************************************************************************************)
unit UBasicVR;

interface

uses
  Winapi.Windows, Winapi.D3D11, Winapi.DxgiFormat, OVR_CAPI, OVR_CAPI_D3D,
  OVR_ErrorCode, OVR_CAPI_Keys, OVR_CAPI_Util, System.SysUtils, System.Math.Vectors,
  UDirectXAppUtil;

type
  TOculusTexture = class
  public
    Session: TOvrSession;
    TextureChain: TOvrTextureSwapChain;
    TexRtv: array of ID3D11RenderTargetView;
    SizeW, SizeH: Integer;
    destructor Destroy; override;
    function Init(session: TOvrSession; sizeW, sizeH: Integer): Boolean;
    function GetRTV: ID3D11RenderTargetView;
    procedure Commit;
  end;

  PQuaternion3D = ^TQuaternion3D;
  TVRLayer = class
  private
    ovrLayer: TOvrLayerEyeFov;
  public
    Session: TOvrSession;
    EyeRenderDesc: array[0..1] of TOvrEyeRenderDesc;     // Description of the VR.
    EyeRenderViewport: array[0..1] of TOvrRecti;         // Useful to remember when varying resolution
    pEyeRenderTexture: array[0..1] of TOculusTexture;    // Where the eye buffers will be rendered
    pEyeDepthBuffer: array[0..1] of TDepthBuffer;        // For the eye buffers to use when rendered
    EyeRenderPose: array[0..1] of TOvrPosef;
    constructor Create(session: TOvrSession; const fov: POvrFovPort = nil;
      pixelsPerDisplayPixel: Single = 1.0);
    destructor Destroy; override;
    procedure MakeEyeBuffers(pixelsPerDisplayPixel: Single = 1.0);
    procedure ConfigureRendering(fov: POvrFovPort = nil);
    function GetEyePoses(useEyeRenderPose: POutEyePoses = nil;
      scaleIPD: PSingle = nil; newIPD: PSingle = nil): TOvrTrackingState;
    function RenderSceneToEyeBuffer(player: TCamera; sceneToRender: TScene;
      eye: Integer; rtv: ID3D11RenderTargetView = nil; eyeRenderPose: POvrPosef = nil;
      timesToRenderRoom: Integer = 1; alpha: Single = 1; red: Single = 1;
      green: Single = 1; blue: Single = 1; nearZ: Single = 0.2; farZ: Single = 1000.0;
      doWeSetupRender: Boolean = True; depthBuffer: TDepthBuffer = nil;
      backRed: Single = 0.0; backGre: Single = 0.0; backBlu: Single = 0.0): TMatrix3D;
    procedure PrepareLayerHeader(leftEyeTexture: TOculusTexture = nil;
      leftEyePose: POvrPosef = nil; extraQuat: PQuaternion3D = nil);
  end;

  TBasicVR = class
  protected
  var
    presentResult: TOvrResult;
    initialized: Boolean;
    FRestart: Boolean;
    mirrorTexture: TOvrMirrorTexture;
    scaleWindowW: Single;
    scaleWindowH: Single;
    scaleMirrorW: Single;
    scaleMirrorH: Single;
    windowed: Boolean;
    Yaw: Single;
    cubeClock: Single;
    function Release: Integer;
  public
    Session: TOvrSession;
    HmdDesc: TOvrHmdDesc;
    TrackerDescArray: array of TOvrTrackerDesc;
    Layer: array[0..Ord(ovrMaxLayerCount)-1] of TVRLayer;
    MainCam: TCamera;
    RoomScene: TScene;
    constructor Create(const title: string = 'BasicVR');
    destructor Destroy; override;
    function Init(retryCreate: Boolean): Boolean;
    function HandleMessages: Boolean;
    procedure MainLoop; virtual; abstract;
    function Run: Integer;
    procedure Restart;
    procedure ActionFromInput(speed: Single = 1.0; updateYaw: Boolean = True;
      verticalControl: Boolean = False);
    procedure DistortAndPresent(numLayersToRender: Integer;
      optionalBoxForMirrorWindow: PD3D11_BOX = nil; mirror: Boolean = True);
  end;

implementation

uses
  System.Math, UCommon;

{ TOculusTexture }

function TOculusTexture.Init(session: TOvrSession; sizeW, sizeH: Integer): Boolean;
var
  desc: TOvrTextureSwapChainDesc;
  ret: TOvrResult;
  textureCount: Int32;
  i: NativeInt;
  tex: ID3D11Texture2D;
  rtvd: D3D11_RENDER_TARGET_VIEW_DESC;
begin
  Self.Session := session;
  Self.SizeW := sizeW;
  Self.SizeH := sizeH;

  desc := Default(TOvrTextureSwapChainDesc);
  desc.Type_ := ovrTexture_2D;
  desc.ArraySize := 1;
  desc.Format := OVR_FORMAT_R8G8B8A8_UNORM_SRGB;
  desc.Width := sizeW;
  desc.Height := sizeH;
  desc.MipLevels := 1;
  desc.SampleCount := 1;
  desc.StaticImage := ovrFalse;
  desc.MiscFlags := Ord(ovrTextureMisc_DX_Typeless);
  desc.BindFlags := Ord(ovrTextureBind_DX_RenderTarget);

  ret := ovr_CreateTextureSwapChainDX(Session, DIRECTX.Device, desc, TextureChain);
  if not OVR_SUCCESS(ret) then
    Exit(False);

  textureCount := 0;
  ovr_GetTextureSwapChainLength(Session, TextureChain, textureCount);
  SetLength(TexRtv, textureCount);
  for i := 0 to textureCount-1 do
  begin
    ovr_GetTextureSwapChainBufferDX(Session, TextureChain, i, ID3D11Texture2D, tex);
    rtvd := Default(D3D11_RENDER_TARGET_VIEW_DESC);
    rtvd.Format := DXGI_FORMAT_R8G8B8A8_UNORM;
    rtvd.ViewDimension := D3D11_RTV_DIMENSION_TEXTURE2D;
    DIRECTX.Device.CreateRenderTargetView(tex, @rtvd, TexRtv[i]);
    tex := nil;
  end;

  Result := True;
end;

destructor TOculusTexture.Destroy;
begin
  if TextureChain <> nil then
  begin
    ovr_DestroyTextureSwapChain(Session, TextureChain);
    TextureChain := nil;
  end;

  inherited;
end;

function TOculusTexture.GetRTV: ID3D11RenderTargetView;
var
  currentIndex: Int32;
begin
  if TextureChain <> nil then
  begin
    currentIndex := 0;
    ovr_GetTextureSwapChainCurrentIndex(Session, TextureChain, currentIndex);
    Exit(TexRtv[currentIndex]);
  end;
  Result := nil;
end;

procedure TOculusTexture.Commit;
begin
  ovr_CommitTextureSwapChain(Session, TextureChain);
end;

{ TVRLayer }

constructor TVRLayer.Create(session: TOvrSession; const fov: POvrFovPort = nil;
  pixelsPerDisplayPixel: Single = 1.0);
begin
  inherited Create;

  Self.Session := session;
  MakeEyeBuffers(pixelsPerDisplayPixel);
  ConfigureRendering(fov);
end;

destructor TVRLayer.Destroy;
var
  eye: NativeInt;
begin
  for eye := 0 to 1 do
  begin
    FreeAndNil(pEyeRenderTexture[eye]);
    FreeAndNil(pEyeDepthBuffer[eye]);
  end;

  inherited;
end;

//-----------------------------------------------------------------------
procedure TVRLayer.MakeEyeBuffers(pixelsPerDisplayPixel: Single = 1.0);
var
  eye: NativeInt;
  idealSize: TOvrSizei;
begin
  for eye := 0 to 1 do
  begin
    idealSize := ovr_GetFovTextureSize(Session, TOvrEyeType(eye), ovr_GetHmdDesc(Session).DefaultEyeFov[eye], pixelsPerDisplayPixel);
    pEyeRenderTexture[eye] := TOculusTexture.Create;
    if not pEyeRenderTexture[eye].Init(Session, idealSize.w, idealSize.h) then
      Exit;
    pEyeDepthBuffer[eye] := TDepthBuffer.Create(DIRECTX.Device, idealSize.w, idealSize.h);
    EyeRenderViewport[eye].Pos.x := 0;
    EyeRenderViewport[eye].Pos.y := 0;
    EyeRenderViewport[eye].Size := idealSize;
  end;
end;

//--------------------------------------------------------
procedure TVRLayer.ConfigureRendering(fov: POvrFovPort = nil);
begin
  // If any values are passed as NULL, then we use the default basic case
  if fov <> nil then
  begin
    EyeRenderDesc[0] := ovr_GetRenderDesc(Session, ovrEye_Left, fov^);
    Inc(fov);
    EyeRenderDesc[1] := ovr_GetRenderDesc(Session, ovrEye_Right, fov^);
  end
  else begin
    EyeRenderDesc[0] := ovr_GetRenderDesc(Session, ovrEye_Left, ovr_GetHmdDesc(Session).DefaultEyeFov[0]);
    EyeRenderDesc[1] := ovr_GetRenderDesc(Session, ovrEye_Right, ovr_GetHmdDesc(Session).DefaultEyeFov[1]);
  end;
end;

//------------------------------------------------------------
function TVRLayer.GetEyePoses(useEyeRenderPose: POutEyePoses = nil;
  scaleIPD: PSingle = nil; newIPD: PSingle = nil): TOvrTrackingState;
var
  useHmdToEyeOffset: THmdToEyeOffset;
  ftiming: Double;
  trackingState: TOvrTrackingState;
begin
  // Get both eye poses simultaneously, with IPD offset already included.
  useHmdToEyeOffset[0] := EyeRenderDesc[0].HmdToEyeOffset;
  useHmdToEyeOffset[1] := EyeRenderDesc[1].HmdToEyeOffset;

  // If any values are passed as NULL, then we use the default basic case
  if useEyeRenderPose = nil then useEyeRenderPose := @EyeRenderPose;
  if scaleIPD <> nil then
  begin
    useHmdToEyeOffset[0].x := useHmdToEyeOffset[0].x * scaleIPD^;
    useHmdToEyeOffset[1].x := useHmdToEyeOffset[1].x * scaleIPD^;
  end;
  if newIPD <> nil then
  begin
    useHmdToEyeOffset[0].x := -(newIPD^ * 0.5);
    useHmdToEyeOffset[1].x := +(newIPD^ * 0.5);
  end;

  ftiming := ovr_GetPredictedDisplayTime(Session, 0);
  trackingState := ovr_GetTrackingState(Session, ftiming, ovrTrue);

  ovr_CalcEyePoses(trackingState.HeadPose.ThePose, useHmdToEyeOffset, useEyeRenderPose^);

  Result := trackingState;
end;

//-----------------------------------------------------------
function TVRLayer.RenderSceneToEyeBuffer(player: TCamera; sceneToRender: TScene;
  eye: Integer; rtv: ID3D11RenderTargetView = nil; eyeRenderPose: POvrPosef = nil;
  timesToRenderRoom: Integer = 1; alpha: Single = 1; red: Single = 1;
  green: Single = 1; blue: Single = 1; nearZ: Single = 0.2; farZ: Single = 1000.0;
  doWeSetupRender: Boolean = True; depthBuffer: TDepthBuffer = nil;
  backRed: Single = 0.0; backGre: Single = 0.0; backBlu: Single = 0.0): TMatrix3D;
var
  needsCommit: Boolean;
  CombinedPos: TPoint3D;
  finalCam: TCamera;
  view, prod: TMatrix3D;
  p: TOvrMatrix4f;
  n: NativeInt;
begin
  // If any values are passed as NULL, then we use the default basic case
  if depthBuffer = nil then depthBuffer := pEyeDepthBuffer[eye];
  if eyeRenderPose = nil then eyeRenderPose := @Self.EyeRenderPose[eye];

  needsCommit := False;
  if doWeSetupRender then
  begin
    // If none specified, then using special, and default, Oculus eye buffer render target
    // Note that otherwise we might be using just an ordinary texture, so can't 
    // always assume its an Oculus texture chain.
    if rtv <> nil then
    begin
      DIRECTX.SetAndClearRenderTarget(rtv, depthBuffer, backRed, backGre, backBlu);
    end
    else begin
      needsCommit := True;
      DIRECTX.SetAndClearRenderTarget(pEyeRenderTexture[eye].GetRTV, depthBuffer, backRed, backGre, backBlu);
    end;

    DIRECTX.SetViewport(EyeRenderViewport[eye].Pos.x, EyeRenderViewport[eye].Pos.y,
      EyeRenderViewport[eye].Size.w, EyeRenderViewport[eye].Size.h);
  end;

  // Get view and projection matrices for the Rift camera
  CombinedPos := player.Pos + TPoint3D(eyeRenderPose.Position) * TMatrix3D(player.Rot);
  finalCam := TCamera.Create(CombinedPos, player.Rot * TQuaternion3D(eyeRenderPose.Orientation));
  try
    view := finalCam.GetViewMatrix;
    p := ovrMatrix4f_Projection(EyeRenderDesc[eye].Fov, nearZ, farZ, Ord(ovrProjection_None));
    prod := view * TMatrix3D(p).Transpose;
  finally
    finalCam.Free;
  end;

  // Render the scene
  for n := 0 to timesToRenderRoom-1 do
    sceneToRender.Render(prod, red, green, blue, alpha, True);

  if needsCommit then
  begin
    // Commit rendering to the swap chain so it can be picked up by the SDK
    pEyeRenderTexture[eye].Commit;
  end;
  Result := prod;
end;

//------------------------------------------------------------
procedure TVRLayer.PrepareLayerHeader(leftEyeTexture: TOculusTexture = nil;
  leftEyePose: POvrPosef = nil; extraQuat: PQuaternion3D = nil);
var
  useEyeTexture: array[0..1] of TOculusTexture;
  useEyeRenderPose: array[0..1] of TOvrPosef;
  i: NativeInt;
  localPoseQuat: TQuaternion3D;
  temp: TQuaternion3D;
begin
  // Use defaults where none specified
  useEyeTexture[0] := pEyeRenderTexture[0];
  useEyeTexture[1] := pEyeRenderTexture[1];
  useEyeRenderPose[0] := EyeRenderPose[0];
  useEyeRenderPose[1] := EyeRenderPose[1];
  if leftEyeTexture <> nil then useEyeTexture[0] := leftEyeTexture;
  if leftEyePose <> nil then useEyeRenderPose[0] := leftEyePose^;

  // If we need to fold in extra rotations to the timewarp, per eye
  // We make the changes to the temporary copy, rather than
  // the global one.
  if extraQuat <> nil then
  begin
    for i := 0 to 1 do
    begin
      localPoseQuat := TQuaternion3D(useEyeRenderPose[i].Orientation);
      temp := localPoseQuat * QuaternionInverse(extraQuat^);
      useEyeRenderPose[i].Orientation.w := temp.RealPart;
      useEyeRenderPose[i].Orientation.x := temp.ImagPart.X;
      useEyeRenderPose[i].Orientation.y := temp.ImagPart.Y;
      useEyeRenderPose[i].Orientation.z := temp.ImagPart.Z;
      Inc(extraQuat);
    end;
  end;

  ovrLayer.Header.Type_ := ovrLayerType_EyeFov;
  ovrLayer.Header.Flags := 0;
  ovrLayer.ColorTexture[0] := useEyeTexture[0].TextureChain;
  ovrLayer.ColorTexture[1] := useEyeTexture[1].TextureChain;
  ovrLayer.RenderPose[0] := useEyeRenderPose[0];
  ovrLayer.RenderPose[1] := useEyeRenderPose[1];
  ovrLayer.Fov[0] := EyeRenderDesc[0].Fov;
  ovrLayer.Fov[1] := EyeRenderDesc[1].Fov;
  ovrLayer.Viewport[0] := EyeRenderViewport[0];
  ovrLayer.Viewport[1] := EyeRenderViewport[1];
end;

{ TBasicVR }

constructor TBasicVR.Create(const title: string = 'BasicVR');
var
  ret: TOvrResult;
begin
  inherited Create;

  presentResult := Ord(ovrSuccess);
  scaleWindowW := 0.5;
  scaleWindowH := 0.5;
  scaleMirrorW := 1.0;
  scaleMirrorH := 1.0;
  windowed := True;

  // Initializes LibOVR, and the Rift
  ret := ovr_Initialize(nil);
  VALIDATE(OVR_SUCCESS(ret), 'Failed to initialize libOVR.');

  VALIDATE(DIRECTX.InitWindow(title), 'Failed to open window.');
end;

destructor TBasicVR.Destroy;
begin
  Release;
  ovr_Shutdown;
  DIRECTX.CloseWindow;

  inherited;
end;

function TBasicVR.Init(retryCreate: Boolean): Boolean;
var
  trackerCount: UInt32;
  i: NativeInt;
  desc: TOvrMirrorTextureDesc;
  luid: TOvrGraphicsLuid;
  ret: TOvrResult;
begin
  ret := ovr_Create(Session, luid);
  if not OVR_SUCCESS(ret) then
    Exit(retryCreate);

  HmdDesc := ovr_GetHmdDesc(Session);

  trackerCount := Max(1, ovr_GetTrackerCount(Session)); // Make sure there's always at least one.
  SetLength(TrackerDescArray, trackerCount);
  for i := 0 to trackerCount-1 do
    TrackerDescArray[i] := ovr_GetTrackerDesc(Session, i);

  // Setup Device and Graphics
  if not DIRECTX.InitDevice(Trunc(scaleWindowW * HmdDesc.Resolution.w),
         Trunc(scaleWindowH * HmdDesc.Resolution.h), luid, windowed) then
    Exit(retryCreate);

  // Create a mirror, to see Rift output on a monitor
  mirrorTexture := nil;
  desc := Default(TOvrMirrorTextureDesc);
  desc.Format := OVR_FORMAT_R8G8B8A8_UNORM_SRGB;
  desc.Width := Trunc(scaleMirrorW * DIRECTX.WinSizeW);
  desc.Height := Trunc(scaleMirrorH * DIRECTX.WinSizeH);
  ret := ovr_CreateMirrorTextureDX(Session, DIRECTX.Device, desc, mirrorTexture);
  if not OVR_SUCCESS(ret) then
  begin
    if retryCreate then
      Exit(True);
    VALIDATE(False, 'Failed to create mirror texture.');
  end;

  // Create the room model in-place
  RoomScene := TScene.Create(False);

  // Create camera in-place
  MainCam := TCamera.Create(Point3D(0.0, 1.6, 5.0), TQuaternion3D.Identity);

  initialized := True;
  Result := initialized;
end;

function TBasicVR.HandleMessages: Boolean;
begin
  // continue MainLoop while SubmitFrame is successful,
  // the use hasn't quit and we're not restarting...
  Result := OVR_SUCCESS(presentResult) and DIRECTX.HandleMessages and not FRestart;
end;

function TBasicVR.Run: Integer;
begin
  // false => exit on any failure
  VALIDATE(Init(False), 'Oculus Rift not detected.');

  while DIRECTX.HandleMessages do
  begin
    if initialized then
    begin
      presentResult := Ord(ovrSuccess);
      MainLoop;
      Release;
    end;

    if not FRestart then
    begin
      if presentResult <> Ord(ovrError_DisplayLost) then
        Break;
      Sleep(10);
    end
    else begin
      FRestart := False;
    end;

    // true => retry on ovrError_DisplayLost
    if not Init(True) then
      Break;
  end;

  OVR_DEBUG_LOG('Run: Done');
  Result := Release;
end;

// Called to reinitialize under the app's control
procedure TBasicVR.Restart;
begin
  FRestart := True;
end;

//-------------------------------------------------------
procedure TBasicVR.ActionFromInput(speed: Single = 1.0; updateYaw: Boolean = True;
  verticalControl: Boolean = False);
var
  fwd, right: TPoint3D;
begin
  // Keyboard inputs to adjust player orientation, unaffected by speed
  if updateYaw then
  begin
    if DIRECTX.Key[Chr(VK_LEFT)] or DIRECTX.Key['Q'] then
    begin
      Yaw := Yaw + 0.02;
      MainCam.Rot := TQuaternion3D.Create(Yaw, 0, 0);
    end;
    if DIRECTX.Key[Chr(VK_RIGHT)] or DIRECTX.Key['E'] then
    begin
      Yaw := Yaw - 0.02;
      MainCam.Rot := TQuaternion3D.Create(Yaw, 0, 0);
    end;
  end;
  // Keyboard inputs to adjust player position
  fwd := Point3D(0, 0, -0.05*speed) * TMatrix3D(MainCam.Rot);
  right := Point3D(0.05*speed, 0, 0) * TMatrix3D(MainCam.Rot);
  if DIRECTX.Key['W'] then MainCam.Pos := MainCam.Pos + fwd;
  if DIRECTX.Key['S'] then MainCam.Pos := MainCam.Pos - fwd;
  if DIRECTX.Key['D'] then MainCam.Pos := MainCam.Pos + right;
  if DIRECTX.Key['A'] then MainCam.Pos := MainCam.Pos - right;
  if verticalControl then
  begin
    if DIRECTX.Key['X']   then MainCam.Pos := MainCam.Pos + Point3D(0, +0.05*speed, 0);
    if DIRECTX.Key['Z'] then MainCam.Pos := MainCam.Pos + Point3D(0, -0.05*speed, 0);
  end
  else begin
    // Set camera height as fixed at this point - might be scaled later
    // It no longer seems to find the preset user height.
    MainCam.Pos := Point3D(MainCam.Pos.X, ovr_GetFloat(Session, OVR_KEY_EYE_HEIGHT, 0), MainCam.Pos.Z);
  end;

  // Animate the cube
  cubeClock := cubeClock + 0.015;
  RoomScene.Models[0].Pos := Point3D(9 * Sin(cubeClock), 3, 9 * Cos(cubeClock));
end;

//------------------------------------------------------------
procedure TBasicVR.DistortAndPresent(numLayersToRender: Integer;
  optionalBoxForMirrorWindow: PD3D11_BOX = nil; mirror: Boolean = True);
var
  layerHeaders: array[0..Ord(ovrMaxLayerCount)-1] of POvrLayerHeader;
  i: NativeInt;
  res: ID3D11Resource;
begin
  for i := 0 to Ord(ovrMaxLayerCount)-1 do
  begin
    if Layer[i] <> nil then
      layerHeaders[i] := @Layer[i].ovrLayer.Header;
  end;

  presentResult := ovr_SubmitFrame(Session, 0, nil, @layerHeaders, numLayersToRender);
  if not OVR_SUCCESS(presentResult) then
    Exit;

  if mirror and (mirrorTexture <> nil) then
  begin
    // Render mirror
    ovr_GetMirrorTextureBufferDX(Session, mirrorTexture, ID3D11Resource, res);
    DIRECTX.Context.CopySubresourceRegion(DIRECTX.BackBuffer, 0, 0, 0, 0, res, 0, optionalBoxForMirrorWindow);
    res := nil;
    DIRECTX.SwapChain.Present(0, 0);
  end;
end;

//------------------------------------------------------------
function TBasicVR.Release: Integer;
var
  i: NativeInt;
begin
  initialized := False;

  if mirrorTexture <> nil then
  begin
    ovr_DestroyMirrorTexture(Session, mirrorTexture);
    mirrorTexture := nil;
  end;

  for i := 0 to Ord(ovrMaxLayerCount)-1 do
  begin
    FreeAndNil(Layer[i]);
  end;

  FreeAndNil(RoomScene);

  DIRECTX.ReleaseDevice;

  if Session <> nil then
  begin
    ovr_Destroy(Session);
    Session := nil;
  end;

  if DIRECTX.Key['Q'] and DIRECTX.Key[Chr(VK_CONTROL)] then
    Exit(99);  // Special return code for quitting sample 99
  Result := 0;
end;

end.