(************************************************************************************
Filename    :   Win32_ControlMethods.h
Content     :   Shared functionality for the VR control methods
Created     :   October 20th, 2014
Author      :   Tom Heath
Copyright   :   Copyright 2014 Oculus, Inc. All Rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*************************************************************************************)
unit UControlMethods;

interface

uses
  System.Math, System.Math.Vectors, OVR_CAPI, OVR_CAPI_Keys, UDirectXAppUtil, UBasicVR;

function GetEulerAngles(const q: TQuaternion3D): TPoint3D;
function GetAutoYawRotation(vrLayer: TVRLayer): TQuaternion3D;
function WasItTapped(const linearAcc: TOvrVector3f): Boolean;
function GetAccelJumpPosY(pBasicVR: TBasicVR; const [Ref] pTrackingState: TOvrTrackingState): Single;
function FindVelocityFromTilt(pBasicVR: TBasicVR; vrLayer: TVRLayer; const [Ref] pTrackingState: TOvrTrackingState): TPoint3D;

implementation

//-------------------------------------------
function GetEulerAngles(const q: TQuaternion3D): TPoint3D;
var
  euler, fwd: TPoint3D;
  horizLength: Single;
  right: TPoint3D;
begin
  fwd := Point3D(0, 0, 1) * TMatrix3D(q);
  euler.Y := ArcTan2(fwd.x , fwd.z);
  horizLength := Sqrt(fwd.x*fwd.x + fwd.z*fwd.z);
  euler.x := ArcTan2(fwd.y, horizLength);
  right := Point3D(1, 0, 0) * TMatrix3D(q);
  horizLength := Sqrt(right.x*right.x + right.z*right.z);
  euler.z := ArcTan2(right.y, horizLength);
  //Util.Output("Pitch = %0.0f, Yaw = %0.0f, Roll = %0.0f\n", XMConvertToDegrees(euler.x), XMConvertToDegrees(euler.y), XMConvertToDegrees(euler.z));
  Result := euler;
end;

//--------------------------------------------
var
  FYaw: Single = 3.141;
function GetAutoYawRotation(vrLayer: TVRLayer): TQuaternion3D;
var
  orientQuat: TQuaternion3D;
  yaw: Single;
begin
  // Increments in yaw are proportional to Rift yaw
  orientQuat := TQuaternion3D(vrLayer.EyeRenderPose[0].Orientation);
  yaw := GetEulerAngles(orientQuat).Y;
  FYaw := FYaw + yaw * 0.02;
  Result := TQuaternion3D.Create(FYaw, 0, 0);
end;

//----------------------------------------------------------------------
var
  FReadyForNewSingleTap: Boolean = False;
function WasItTapped(const linearAcc: TOvrVector3f): Boolean;
const
  thresholdForTap = 10.0;
  thresholdForReset = 2.0;
var
  magOfAccel: Single;
begin
  magOfAccel := TPoint3D(linearAcc).Length;
  if magOfAccel < thresholdForReset then
    FReadyForNewSingleTap := True;
  if FReadyForNewSingleTap and (magOfAccel > thresholdForTap) then
  begin
    FReadyForNewSingleTap := False;
    Exit(True);
  end;
  Result := False;
end;

//---------------------------------------------------
var
  FIsStaticVarInitialized: Boolean = False;
  yPos: Single;
  yVel: Single = 0;
function GetAccelJumpPosY(pBasicVR: TBasicVR; const [Ref] pTrackingState: TOvrTrackingState): Single;
const
  jumpResponse = 0.0015;
begin
  // Change y position - note, we need to keep a version here
  // As the one in ActionFromInput will keep resetting us to height of the character
  if not FIsStaticVarInitialized then
  begin
    yPos := ovr_GetFloat(pBasicVR.Session, OVR_KEY_EYE_HEIGHT, 0);
    FIsStaticVarInitialized := True;
  end;

  // Jump into air manually with '1' key
  if DIRECTX.Key[' '] then yVel := yVel + 0.01;

  // Jump from head movement - note, its slightly counter intuitive - instead of responding to the 'up' accel,
  // it lets that one get absorbed by the ground, and acts on the deceleration from the upward movement.
  // Alternative : You could also try yVel += - min(0,jumpResponse*pTrackingState->HeadPose.LinearAcceleration.y);
  //               in the below code. This has facets that are interesting, in allowing multiple jumps,
  //               and eliminates any downward midair movement - but leaves the motion unbounded.
  yVel := yVel - jumpResponse * pTrackingState.HeadPose.LinearAcceleration.y;

  // Add pseudo gravity
  yVel := yVel - 0.002;

  // Increment position
  yPos := yPos + yVel;

  // Hit floor;
  if yPos < 1.6 then ///ovr_GetFloat(pBasicVR.Session, OVR_KEY_EYE_HEIGHT, 0))  No longer works
  begin
    yPos := 1.6; /// ovr_GetFloat(pBasicVR.Session, OVR_KEY_EYE_HEIGHT, 0);
    yVel := 0;
  end;

  Result := yPos;
end;

//-----------------------------------------------------------------------
var
  vel: TPoint3D;
function FindVelocityFromTilt(pBasicVR: TBasicVR; vrLayer: TVRLayer; const [Ref] pTrackingState: TOvrTrackingState): TPoint3D;
const
  tiltResponse = 0.0075;
  maxSpeed = 0.1;
var
  orientQuat: TQuaternion3D;
  eulerFromRift: TPoint3D;
  totalHorizRot: TQuaternion3D;
  unitForwardVector, unitRightVector: TPoint3D;
  speed: Single;
begin
  // Find the orthogonal vectors resulting from combined rift and user yaw
  orientQuat := TQuaternion3D(vrLayer.EyeRenderPose[0].Orientation);
  eulerFromRift := GetEulerAngles(orientQuat);

  totalHorizRot := pBasicVR.MainCam.Rot * TQuaternion3D.Create(eulerFromRift.Y, 0, 0);
  unitForwardVector := Point3D(0,0,-1) * TMatrix3D(totalHorizRot);
  unitRightVector := Point3D(1,0,0) * TMatrix3D(totalHorizRot);

  // Now feed into a static velocity.

  // Hold down space if you want to look around, instead of move
  if not DIRECTX.Key[' '] then
  begin
    vel := vel + unitForwardVector * (tiltResponse * Tan(eulerFromRift.X)); // Pitch
    vel := vel - unitRightVector * (tiltResponse * Tan(eulerFromRift.Z)); // Roll
  end;

  // We always have damping, to cap top speeds
  // and to damp to zero when space is released.
  vel := vel * 0.98;

  // Limit velocity
  speed := vel.Length;
  if speed > maxSpeed then vel := vel * (maxSpeed / speed);

  Result := vel;
end;

end.