unit UDemoZoom;

interface

procedure Main;

implementation

uses
  System.Math, System.Math.Vectors, OVR_CAPI, OVR_CAPI_Util, UDirectXAppUtil, UBasicVR;

type
  TZoomForIO = class(TBasicVR)
  private
    amountOfZoom: Single;
    howFarAway: Single;
  public
    constructor Create;
    procedure MainLoop; override;
  end;

{ TZoomForIO }

constructor TZoomForIO.Create;
begin
  inherited Create('Zoom For IO');

  amountOfZoom := 0.1;
  howFarAway := 0.75;
end;

procedure TZoomForIO.MainLoop;
var
  zoomedTexture: TTexture;
  scopeScale: Single;
  cube: TTriangleSet;
  sniperModel: TModel;
  blackBorder: Integer;
  eyeQuat: TQuaternion3D;
  eyePos, CombinedPos: TPoint3D;
  finalCam: TCamera;
  view, proj, projView: TMatrix3D;
  zoomedFOV: TOvrFovPort;
  eye: NativeInt;
  StaticMainCam: TCamera;
begin
  Layer[0] := TVRLayer.Create(Session);

  // Make a texture to render the zoomed image into.  Make it same size as left eye buffer, for simplicity.
  zoomedTexture := TTexture.Create(True, Max(Layer[0].pEyeRenderTexture[0].SizeW, Layer[0].pEyeRenderTexture[1].SizeW),
                                         Max(Layer[0].pEyeRenderTexture[0].SizeH, Layer[0].pEyeRenderTexture[1].SizeH));

  // Make a scope model - its small and close to us
  scopeScale := 0.25;
  cube := Default(TTriangleSet);
  cube.AddQuad(TVertex.Create(Point3D( scopeScale,  scopeScale, 0), $FFFFFFFF, 0, 0),
               TVertex.Create(Point3D(-scopeScale,  scopeScale, 0), $FFFFFFFF, 1, 0),
               TVertex.Create(Point3D( scopeScale, -scopeScale, 0), $FFFFFFFF, 0, 1),
               TVertex.Create(Point3D(-scopeScale, -scopeScale, 0), $FFFFFFFF, 1, 1));
  sniperModel := TModel.Create(cube, Point3D(0, 0, 0), TQuaternion3D.Identity, TMaterial.Create(zoomedTexture));

  while HandleMessages do
  begin
    ActionFromInput;
    Layer[0].GetEyePoses;

    // Render the zoomed scene, making sure we clear the back screen with solid alpha
    DIRECTX.SetAndClearRenderTarget(zoomedTexture.TexRtv, Layer[0].pEyeDepthBuffer[0], 0, 0, 0, 1);

    // Lets set a slightly small viewport, so we get a black border
    blackBorder := 16;
    DIRECTX.SetViewport(Layer[0].EyeRenderViewport[0].Pos.x + blackBorder,
                        Layer[0].EyeRenderViewport[0].Pos.y + blackBorder,
                        Layer[0].EyeRenderViewport[0].Size.w - 2*blackBorder,
                        Layer[0].EyeRenderViewport[0].Size.h - 2*blackBorder);

    //Get the pose information in XM format
    eyeQuat := TQuaternion3D(Layer[0].EyeRenderPose[0].Orientation);
    eyePos  := TPoint3D(Layer[0].EyeRenderPose[0].Position);

    // Get view and projection matrices for the Rift camera
    CombinedPos := MainCam.Pos + eyePos * TMatrix3D(MainCam.Rot);
    finalCam := TCamera.Create(CombinedPos, MainCam.Rot * eyeQuat);
    try
      view := finalCam.GetViewMatrix;
    finally
      finalCam.Free;
    end;

    // Vary amount of zoom with '1' and '2'Lets pick a zoomed in FOV
    if DIRECTX.Key['1'] then amountOfZoom := Max(amountOfZoom - 0.002, 0.050);
    if DIRECTX.Key['2'] then amountOfZoom := Min(amountOfZoom + 0.002, 0.500);
    zoomedFOV.DownTan := amountOfZoom;
    zoomedFOV.UpTan := amountOfZoom;
    zoomedFOV.LeftTan := amountOfZoom;
    zoomedFOV.RightTan := amountOfZoom;

    // Finally, render zoomed scene onto the texture
    proj := TMatrix3D(ovrMatrix4f_Projection(zoomedFOV, 0.2, 1000.0, Ord(ovrProjection_None))).Transpose;
    projView := view * proj;
    RoomScene.Render(projView, 1, 1, 1, 1, True);

    for eye := 0 to 1 do
    begin
      // Render main, outer world
      Layer[0].RenderSceneToEyeBuffer(MainCam, RoomScene, eye);

      // Render scope with special static camera, always in front of us
      if DIRECTX.Key['3'] then howFarAway := max(howFarAway - 0.002, 0.25);
      if DIRECTX.Key['4'] then howFarAway := min(howFarAway + 0.002, 1.00);
      StaticMainCam := TCamera.Create(Point3D(0, 0, -howFarAway), TQuaternion3D.Create(3.14, 0, 0));
      try
        view := StaticMainCam.GetViewMatrix;
      finally
        StaticMainCam.Free;
      end;
      proj := TMatrix3D(ovrMatrix4f_Projection(Layer[0].EyeRenderDesc[eye].Fov, 0.2, 1000.0, Ord(ovrProjection_None))).Transpose;
      projView := view * proj;
      sniperModel.Render(projView, 1, 1, 1, 1, True);
    end;

    Layer[0].PrepareLayerHeader;
    DistortAndPresent(1);
  end;

  sniperModel.Free;
end;

procedure Main;
var
  app: TZoomForIO;
begin
  app := TZoomForIO.Create;
  try
    ExitCode := app.Run;
  finally
    app.Free;
  end;
end;

end.